#!/usr/bin/env python

'''
          Pysr - A small python relative searcher
       Copyright (C) 2008  TkDrg
            
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or 
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''

from array import array
from glob import glob
from optparse import OptionParser
import os.path
from sets import Set
import sys

__version__ = "0.01"

charset = "ABDCEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    option, filenames, search_str, parser = ParseArgs(argv)

    filenames = Set(filenames)

    # If psyco is installed, use it.
    try:
        import psyco
    except:
        pass
    else:
        psyco.full()

    pattern = MakePattern(search_str, option, parser)
    sstrlen = len(pattern)+1

    for fname in filenames:
        SearchFile(fname, pattern, sstrlen, option)

    return 0

def SearchFile(filename, pattern, sstrlen, option):
    chunk_size = int("1024")

    # if string has no skips, it can be optimized
    if None in pattern:
        searcher = ComplexSearch
    else:
        searcher = SimpleSearch

    # make the search
    try:
        searchfile = file(filename, 'rb')
    except:
        value = sys.exc_info()[1]
        print >> sys.stderr, "Error reading %s; ignoring\n\t%s" % \
            (filename, value)
        return

    try:
        eof = False
        while not eof:
            chunk_offset = searchfile.tell()
            raw_chunk = searchfile.read(chunk_size)

            if raw_chunk == '':
                break
            elif len(raw_chunk) < chunk_size:
                eof = True

            searcher(filename, raw_chunk, pattern, chunk_offset)
            searchfile.seek(-sstrlen+1, 1)
    finally:
        searchfile.close()

def SimpleSearch(filename, raw_chunk, pattern, chunk_offset):
    chunk = PreprocessChunk(raw_chunk)
    length = len(pattern)
    first = pattern[0]
    pos = 0
    while 1:
        try:
            index = chunk.index(first, pos)
        except ValueError:
            break
        else:
            end = index+length
            if chunk[index:end] == pattern:
                offset = chunk_offset+index
                PrintMatch(filename, offset, raw_chunk[index:end+1])
            pos = index+1

def ComplexSearch(filename, raw_chunk, pattern, chunk_offset):
    chunk = PreprocessChunk(raw_chunk)
    length = len(pattern)

    skips = []
    for x in range(len(pattern)):
        if pattern[x] is None:
            skips.append(x)

    pos = 0
    done = False
    while not done:
        end = pos+length
        if end == len(chunk):
            done = True
        if MatchComplexPattern(chunk[pos:end], pattern, skips):
            PrintMatch(filename, chunk_offset+pos, raw_chunk[pos:end+1])
        pos += 1

def PrintMatch(filename, offset, seq):
    match = ''
    for x in seq:
        match = "%s%02x " % (match, ord(x))
    match = match[:-1]

    # Display matches
    print "%s (%08x): %s" % (filename, offset, match)

def MatchComplexPattern(data, pattern, skips):
    pattern = pattern[:]
    for x in skips:
        pattern[x] = data[x]
        pattern[x+1] -= data[x]
    return data == pattern

def MakePattern(search_str, option, parser):
    if option.hex:
        bytes = search_str.split()
        if len(bytes) < 2:
            parser.error("Search string too short")

        for x in range(len(bytes)):
            byte = bytes[x]
            if byte in ('*', '**'):
                # skipped byte
                bytes[x] = None
                continue
            if len(byte) > 2:
                parser.error("Invalid search string supplied")
            try:
                bytes[x] = int(byte, 16)
            except ValueError:
                parser.error("Invalid search string supplied")

        pattern = []
        previous = bytes[0]
        for b in bytes[1:]:
            if b is None:
                pattern.append(None)
            else:
                pattern.append(b - previous)
                previous = b
    else:
        if search_str[0] not in charset:
            parser.error("First character of search string must be a letter")
        if search_str[-1] not in charset:
            parser.error("Last character of search string must be a letter")
        if len(search_str) < 2:
            parser.error("Search string too short")

        pattern = []
        previous = charset.find(search_str[0])
        for ch in search_str[1:]:
            ch = charset.find(ch)
            if ch == -1:
                pattern.append(None)    # skipped char
            else:
                pattern.append(ch - previous)
                previous = ch

    return pattern

def PreprocessChunk(raw_chunk):
    raw_chunk = array('B', raw_chunk).tolist()
    chunk_len = len(raw_chunk)-1
    chunk = [0]*chunk_len

    previous = raw_chunk[0]
    for x in range(chunk_len):
        byte = raw_chunk[x+1]
        chunk[x] = byte - previous
        previous = byte
    return chunk

def ParseArgs(argv):
    hex = 'False'

    # set up the option parser
    usage = '''%prog [-x] searchpattern filetosearch [filetosearch2 ...]

    Pysr  Copyright (C) 2008  Tkdrg
    This program comes with ABSOLUTELY NO WARRANTY. This is free software,
    and you are welcome to redistribute it under certain conditions. See
    LICENSE.txt for details.
    '''
    parser = OptionParser(usage=usage)
    parser.add_option('-x', '--hex', action="store_true", dest="hex",)
    option, args = parser.parse_args(argv)

    if len(args) < 2:
        parser.error("Not enough arguments (try -h for help)")

    search_str = args[0]
    filenames = []
    for name in args[1:]:
        result = glob(name)
        if len(result) == 0:
            print >> sys.stderr, name, "not found; ignoring"
        else:
            filenames += result

    return option, filenames, search_str, parser


if __name__ == '__main__':
    sys.exit(main())

